/* SPDX-FileCopyrightText: 2020-2022 The Atlite Authors
   SPDX-License-Identifier: GPL-3.0-or-later
*/

.wy-side-nav-search {
    background-color: #eeeeee;
}

.wy-side-nav-search .wy-dropdown>a,
.wy-side-nav-search>a {
    color: rgb(34, 97, 156)
}

.wy-side-nav-search>div.version {
    color: rgb(34, 97, 156)
}

.wy-menu-vertical header,
.wy-menu-vertical p.caption,
.rst-versions a {
    color: #999999;
}

.wy-menu-vertical a.reference:hover,
.wy-menu-vertical a.reference.internal:hover {
    background: #dddddd;
    color: #fff;
}

.wy-nav-side {
    background: #efefef;
}

.wy-menu-vertical a.reference {
    color: #000;
}

.rst-versions .rst-current-version,
.wy-nav-top,
.wy-menu-vertical li.toctree-l2.current li.toctree-l3>a:hover {
    background: #002221;
}

.wy-nav-content .highlight {
    background: #ffffff;
}

.rst-content code.literal,
.rst-content tt.literal {
    color: rgb(34, 97, 156)
}

.wy-nav-content a.reference {
    color: rgb(34, 97, 156);
}


/* override table width restrictions */

@media screen and (min-width: 767px) {
    .wy-table-responsive table td {
        /* !important prevents the common CSS stylesheets from overriding
          this as on RTD they are loaded after this stylesheet */
        white-space: normal !important;
        background: rgb(250, 250, 250) !important;
    }
    .wy-table-responsive {
        max-width: 100%;
        overflow: visible !important;
    }
    .wy-nav-content {
        max-width: 910px !important;
    }
}