# -*- coding: utf-8 -*-

# SPDX-FileCopyrightText: 2016-2019 The Atlite Authors
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import era5, sarah, gebco

modules  = {'era5': era5, 'sarah': sarah, 'gebco': gebco}